import importlib
import os
import warnings
from ctypes import c_double, c_int
from typing import Optional

import numpy as np
from ase.atoms import Atoms
from ase.calculators.lammps import Prism
from ase.data import atomic_masses, atomic_numbers
from executorlib import BaseExecutor
from scipy import constants

from pylammpsmpi.wrapper.base import LammpsBase


class LammpsASELibrary:
    """
    A class that provides an interface to interact with LAMMPS using ASE.

    Args:
        working_directory (str, optional): The working directory. Defaults to None.
        cores (int, optional): The number of cores to use. Defaults to 1.
        comm (object, optional): The MPI communicator. Defaults to None.
        logger (object, optional): The logger object. Defaults to None.
        log_file (str, optional): The log file path. Defaults to None.
        library (object, optional): The LAMMPS library object. Defaults to None.
        disable_log_file (bool, optional): Whether to disable the log file. Defaults to True.
        executor: Executor to use for parallel execution (default: None)
    """

    def __init__(
        self,
        working_directory: Optional[str] = None,
        hostname_localhost: Optional[bool] = None,
        cores: int = 1,
        comm: Optional[object] = None,
        logger: Optional[object] = None,
        log_file: Optional[str] = None,
        library: Optional[object] = None,
        disable_log_file: bool = True,
        executor: Optional[BaseExecutor] = None,
    ):
        self._logger = logger
        self._prism = None
        self._structure = None
        self._cores = cores
        if disable_log_file and log_file is None:
            cmdargs = ["-screen", "none", "-log", "none"]
        else:
            if log_file is None and working_directory is not None:
                log_file = os.path.join(working_directory, "log.lammps")
            elif log_file is None:
                log_file = "log.lammps"
            cmdargs = ["-screen", "none", "-log", log_file]
        if library is not None:
            self._interactive_library = library
            self._cores = library.cores
        elif self._cores == 1:
            lammps = importlib.import_module("lammps").lammps
            self._interactive_library = lammps(
                cmdargs=cmdargs,
                comm=comm,
            )
        else:
            self._interactive_library = LammpsBase(
                cores=self._cores,
                working_directory=working_directory,
                hostname_localhost=hostname_localhost,
                executor=executor,
                cmdargs=cmdargs,
            )

    def interactive_lib_command(self, command: str) -> None:
        """
        Execute a LAMMPS command using the interactive library.

        Args:
            command (str): The LAMMPS command to execute.
        """
        if self._logger is not None:
            self._logger.debug("Lammps library: " + command)
        self._interactive_library.command(command)

    def interactive_positions_getter(self) -> np.ndarray:
        """
        Get the positions of atoms from the interactive library.

        Returns:
            np.ndarray: The positions of atoms.
        """
        positions = np.reshape(
            np.array(self._interactive_library.gather_atoms("x", 1, 3)),
            (len(self._structure), 3),
        )
        if not _check_ortho_prism(prism=self._prism):
            positions = self._prism.vector_to_ase(positions)
        return positions

    def interactive_positions_setter(self, positions: list[list[float]]) -> None:
        """
        Set the positions of atoms in the interactive library.

        Args:
            positions (List[List[float]]): The positions of atoms.
        """
        positions = _vector_to_lammps(vector=positions, prism=self._prism)
        if self._cores == 1:
            self._interactive_library.scatter_atoms(
                "x", 1, 3, (len(positions) * c_double)(*positions)
            )
        else:
            self._interactive_library.scatter_atoms("x", positions)
        self.interactive_lib_command(command="change_box all remap")

    def interactive_cells_getter(self) -> np.ndarray:
        """
        Get the cell vectors from the interactive library.

        Returns:
            np.ndarray: The cell vectors.
        """
        cc = np.array(
            [
                [self._interactive_library.get_thermo("lx"), 0, 0],
                [
                    self._interactive_library.get_thermo("xy"),
                    self._interactive_library.get_thermo("ly"),
                    0,
                ],
                [
                    self._interactive_library.get_thermo("xz"),
                    self._interactive_library.get_thermo("yz"),
                    self._interactive_library.get_thermo("lz"),
                ],
            ]
        )
        return self._prism.vector_to_ase(cc)

    def interactive_cells_setter(self, cell: np.ndarray) -> None:
        """
        Set the cell vectors in the interactive library.

        Args:
            cell (np.ndarray): The cell vectors.
        """
        self._prism = Prism(cell)
        lx, ly, lz, xy, xz, yz = self._prism.get_lammps_prism()
        if not _check_ortho_prism(prism=self._prism):
            warnings.warn(
                "Warning: setting upper trangular matrix might slow down the calculation",
                stacklevel=2,
            )

        is_skewed = cell_is_skewed(cell=cell, tolerance=1.0e-8)
        was_skewed = cell_is_skewed(cell=self._structure.cell, tolerance=1.0e-8)

        if is_skewed:
            if not was_skewed:
                self.interactive_lib_command(command="change_box all triclinic")
            self.interactive_lib_command(
                command=f"change_box all x final 0 {lx:f} y final 0 {ly:f} z final 0 {lz:f}  xy final {xy:f} xz final {xz:f} yz final {yz:f} remap units box",
            )
        elif was_skewed:
            self.interactive_lib_command(
                command=f"change_box all x final 0 {lx:f} y final 0 {ly:f} z final 0 {lz:f} xy final {0.0:f} xz final {0.0:f} yz final {0.0:f} remap units box",
            )
            self.interactive_lib_command(command="change_box all ortho")
        else:
            self.interactive_lib_command(
                command=f"change_box all x final 0 {lx:f} y final 0 {ly:f} z final 0 {lz:f} remap units box",
            )

    def interactive_volume_getter(self) -> float:
        """
        Get the volume of the system from the interactive library.

        Returns:
            float: The volume of the system.
        """
        return self._interactive_library.get_thermo("vol")

    def interactive_forces_getter(self) -> np.ndarray:
        """
        Get the forces on atoms from the interactive library.

        Returns:
            np.ndarray: The forces on atoms.
        """
        ff = np.reshape(
            np.array(self._interactive_library.gather_atoms("f", 1, 3)),
            (len(self._structure), 3),
        )
        if not _check_ortho_prism(prism=self._prism):
            ff = self._prism.vector_to_ase(ff)
        return ff

    def interactive_structure_setter(
        self,
        structure: Atoms,
        units: str,
        dimension: int,
        boundary: str,
        atom_style: str,
        el_eam_lst: list[str],
        calc_md: bool = True,
    ) -> None:
        """
        Set the structure in the interactive library.

        Args:
            structure (ase.Atoms): The structure.
            units (str): The units of the simulation.
            dimension (int): The dimension of the simulation.
            boundary (str): The boundary conditions.
            atom_style (str): The atom style.
            el_eam_lst (List[str]): The list of element symbols.
            calc_md (bool, optional): Whether to calculate molecular dynamics. Defaults to True.
        """
        self.interactive_lib_command(command="clear")
        control_dict = set_selective_dynamics(structure=structure, calc_md=calc_md)
        self.interactive_lib_command(command="units " + units)
        self.interactive_lib_command(command="dimension " + str(dimension))
        self.interactive_lib_command(command="boundary " + boundary)
        self.interactive_lib_command(command="atom_style " + atom_style)

        self.interactive_lib_command(command="atom_modify map array")
        self._prism = Prism(structure.cell)
        if not _check_ortho_prism(prism=self._prism):
            warnings.warn(
                "Warning: setting upper trangular matrix might slow down the calculation",
                stacklevel=2,
            )
        xhi, yhi, zhi, xy, xz, yz = self._prism.get_lammps_prism()
        if self._prism.is_skewed():
            self.interactive_lib_command(
                command="region 1 prism"
                + " 0.0 "
                + str(xhi)
                + " 0.0 "
                + str(yhi)
                + " 0.0 "
                + str(zhi)
                + " "
                + str(xy)
                + " "
                + str(xz)
                + " "
                + str(yz)
                + " units box",
            )
        else:
            self.interactive_lib_command(
                command="region 1 block"
                + " 0.0 "
                + str(xhi)
                + " 0.0 "
                + str(yhi)
                + " 0.0 "
                + str(zhi)
                + " units box",
            )
        el_struct_lst = get_species_symbols(structure)
        if atom_style == "full":
            self.interactive_lib_command(
                command="create_box "
                + str(len(el_eam_lst))
                + " 1 "
                + "bond/types 1 "
                + "angle/types 1 "
                + "extra/bond/per/atom 2 "
                + "extra/angle/per/atom 2 ",
            )
        else:
            self.interactive_lib_command(
                command="create_box " + str(len(el_eam_lst)) + " 1"
            )
        el_dict = {}
        for id_eam, el_eam in enumerate(el_eam_lst):
            if el_eam in el_struct_lst:
                self.interactive_lib_command(
                    command=f"mass {id_eam + 1:3d} {atomic_masses[atomic_numbers[el_eam]]:f}",
                )
            else:
                self.interactive_lib_command(
                    command=f"mass {id_eam + 1:3d} {1.00:f}",
                )
        positions = _vector_to_lammps(vector=structure.positions, prism=self._prism)
        velocities = _vector_to_lammps(
            vector=structure.get_velocities(),
            prism=self._prism,
        )
        try:
            elem_all = get_lammps_indicies_from_ase_structure(
                structure=structure, el_eam_lst=el_eam_lst
            )
        except KeyError:
            missing = set(get_species_symbols(structure)).difference(el_dict.keys())
            missing = ", ".join([el.Abbreviation for el in missing])
            raise ValueError(
                f"Structure contains elements [{missing}], that are not present in the potential!"
            )
        if self._cores == 1:
            self._interactive_library.create_atoms(
                n=len(structure),
                id=None,
                type=elem_all,
                x=positions,
                v=velocities,
                image=None,
                shrinkexceed=False,
            )
        else:
            self._interactive_library.create_atoms(
                n=len(structure),
                id=range(1, len(structure) + 1),
                type=elem_all,
                x=positions,
                v=velocities,
                image=None,
                shrinkexceed=False,
            )
        self.interactive_lib_command(command="change_box all remap")
        for key, value in control_dict.items():
            self.interactive_lib_command(command=key + " " + value)
        self._structure = structure

    def interactive_indices_getter(self) -> np.ndarray:
        """
        Get the indices of atoms from the interactive library.

        Returns:
            np.ndarray: The indices of atoms.
        """
        return np.array(self._interactive_library.gather_atoms("type", 0, 1))

    def interactive_energy_pot_getter(self) -> float:
        """
        Get the potential energy from the interactive library.

        Returns:
            float: The potential energy.
        """
        return self._interactive_library.get_thermo("pe")

    def interactive_energy_tot_getter(self) -> float:
        """
        Get the total energy from the interactive library.

        Returns:
            float: The total energy.
        """
        return self._interactive_library.get_thermo("etotal")

    def interactive_steps_getter(self) -> int:
        """
        Get the number of steps from the interactive library.

        Returns:
            int: The number of steps.
        """
        return self._interactive_library.get_thermo("step")

    def interactive_temperatures_getter(self) -> float:
        """
        Get the temperature from the interactive library.

        Returns:
            float: The temperature.
        """
        return self._interactive_library.get_thermo("temp")

    def interactive_pressures_getter(self):
        pp = np.array(
            [
                [
                    self._interactive_library.get_thermo("pxx"),
                    self._interactive_library.get_thermo("pxy"),
                    self._interactive_library.get_thermo("pxz"),
                ],
                [
                    self._interactive_library.get_thermo("pxy"),
                    self._interactive_library.get_thermo("pyy"),
                    self._interactive_library.get_thermo("pyz"),
                ],
                [
                    self._interactive_library.get_thermo("pxz"),
                    self._interactive_library.get_thermo("pyz"),
                    self._interactive_library.get_thermo("pzz"),
                ],
            ]
        )
        if not _check_ortho_prism(prism=self._prism):
            pp = self._prism.tensor2_to_ase(pp)
        return pp

    def interactive_indices_setter(self, indices, el_eam_lst):
        elem_all = get_lammps_indicies_from_ase_indices(
            indices=indices, structure=self._structure, el_eam_lst=el_eam_lst
        )
        if self._cores == 1:
            self._interactive_library.scatter_atoms(
                "type", 0, 1, (len(elem_all) * c_int)(*elem_all)
            )
        else:
            self._interactive_library.scatter_atoms("type", elem_all)

    def interactive_stress_getter(self, enable_stress_computation=True):
        """
        This gives back an Nx3x3 array of stress/atom defined in http://lammps.sandia.gov/doc/compute_stress_atom.html
        Keep in mind that it is stress*volume in eV. Further discussion can be found on the website above.

        Returns:
            numpy.array: Nx3x3 np array of stress/atom
        """
        if enable_stress_computation:
            self.interactive_lib_command("compute st all stress/atom NULL")
            self.interactive_lib_command("run 0")
        id_lst = self._interactive_library.extract_atom("id")
        id_lst = np.array([id_lst[i] for i in range(len(self._structure))]) - 1
        id_lst = np.arange(len(id_lst))[np.argsort(id_lst)]
        ind = np.array([0, 3, 4, 3, 1, 5, 4, 5, 2])
        ss = self._interactive_library.extract_compute("st", 1, 2)
        ss = np.array(
            [ss[i][j] for i in range(len(self._structure)) for j in range(6)]
        ).reshape(-1, 6)[id_lst]
        ss = (
            ss[:, ind].reshape(len(self._structure), 3, 3)
            / constants.eV
            * constants.bar
            * constants.angstrom**3
        )
        if not _check_ortho_prism(prism=self._prism):
            ss = np.einsum("ij,njk->nik", self._prism.rot_mat, ss)
            ss = np.einsum("nij,kj->nik", ss, self._prism.rot_mat)
        return ss

    def interactive_velocities_getter(self):
        velocity = np.reshape(
            np.array(self._interactive_library.gather_atoms("v", 1, 3)),
            (len(self._structure), 3),
        )
        if not _check_ortho_prism(prism=self._prism):
            velocity = self._prism.vector_to_ase(velocity)
        return velocity

    def close(self):
        if self._interactive_library is not None:
            self._interactive_library.close()

    def set_fix_external_callback(self, fix_id, callback, caller=None):
        self._interactive_library.set_fix_external_callback(
            fix_id=fix_id, callback=callback, caller=caller
        )

    def __enter__(self):
        """
        Compatibility function for the with statement
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Compatibility function for the with statement
        """
        self.close()


def cell_is_skewed(cell, tolerance=1.0e-8):
    """
    Check whether the simulation box is skewed/sheared. The algorithm compares the box volume
    and the product of the box length in each direction. If these numbers do not match, the box
    is considered to be skewed and the function returns True

    Args:
        tolerance (float): Relative tolerance above which the structure is considered as skewed

    Returns:
        (bool): Whether the box is skewed or not.
    """
    volume = np.abs(np.linalg.det(cell))
    prod = np.linalg.norm(cell, axis=-1).prod()
    return not (volume > 0 and abs(volume - prod) / volume < tolerance)


def _vector_to_lammps(vector, prism):
    if vector is not None:
        if not _check_ortho_prism(prism=prism):
            vector = prism.vector_to_lammps(vector)
        return vector.flatten()
    else:
        return None


def _check_ortho_prism(prism, rtol=0.0, atol=1e-08):
    """
    Check if the rotation matrix of the UnfoldingPrism object is sufficiently close to a unit matrix

    Args:
        prism (ase.calculators.lammps.Prism): Prism object to check
        rtol (float): relative precision for numpy.isclose()
        atol (float): absolute precision for numpy.isclose()

    Returns:
        boolean: True or False
    """
    return np.isclose(prism.rot_mat, np.eye(3), rtol=rtol, atol=atol).all()


def get_species_symbols(structure):
    return np.array(sorted(structure.symbols.indices().keys()))


def get_species_indices_dict(structure):
    return {el: i for i, el in enumerate(sorted(structure.symbols.indices().keys()))}


def get_structure_indices(structure):
    element_indices_dict = get_species_indices_dict(structure=structure)
    elements = np.array(structure.get_chemical_symbols())
    indices = elements.copy()
    for k, v in element_indices_dict.items():
        indices[elements == k] = v
    return indices.astype(int)


def get_lammps_indicies_from_ase_indices(indices, structure, el_eam_lst):
    el_struct_lst = get_species_symbols(structure=structure)
    el_pot_dict = {
        el_eam: id_eam + 1
        for id_eam, el_eam in enumerate(el_eam_lst)
        if el_eam in el_struct_lst
    }
    ind_translate_dict = {
        i: el_pot_dict[el]
        for i, el in enumerate(sorted(structure.symbols.indices().keys()))
    }
    elem_all = indices.copy()
    for k, v in ind_translate_dict.items():
        elem_all[indices == k] = v
    return elem_all.astype(int)


def get_lammps_indicies_from_ase_structure(structure, el_eam_lst):
    el_struct_lst = get_species_symbols(structure=structure)
    el_pot_dict = {
        el_eam: id_eam + 1
        for id_eam, el_eam in enumerate(el_eam_lst)
        if el_eam in el_struct_lst
    }
    symbols_lst = np.array(structure.get_chemical_symbols())
    elem_all = symbols_lst.copy()
    for k, v in el_pot_dict.items():
        elem_all[symbols_lst == k] = v
    return elem_all.astype(int)


def get_fixed_atom_boolean_vector(structure):
    fixed_atom_vector = np.array([[False, False, False]] * len(structure))
    for c in structure.constraints:
        c_dict = c.todict()
        if c_dict["name"] == "FixAtoms":
            fixed_atom_vector[c_dict["kwargs"]["indices"]] = [True, True, True]
        elif c_dict["name"] == "FixedPlane":
            if all(np.isin(c_dict["kwargs"]["direction"], [0, 1, 1 / np.sqrt(2)])):
                if "indices" in c_dict["kwargs"]:
                    fixed_atom_vector[c_dict["kwargs"]["indices"]] = np.array(
                        c_dict["kwargs"]["direction"]
                    ).astype(bool)
                elif "a" in c_dict["kwargs"]:
                    fixed_atom_vector[c_dict["kwargs"]["a"]] = np.array(
                        c_dict["kwargs"]["direction"]
                    ).astype(bool)
            else:
                raise ValueError(
                    "Currently the directions are limited to [1, 0, 0], [1, 1, 0], [1, 1, 1] and its permutations."
                )
        else:
            raise ValueError("Only FixAtoms and FixedPlane are currently supported. ")
    return fixed_atom_vector


def set_selective_dynamics(structure, calc_md):
    control_dict = {}
    if len(structure.constraints) > 0:
        sel_dyn = get_fixed_atom_boolean_vector(structure=structure)
        # Enter loop only if constraints present
        if len(np.argwhere(np.any(sel_dyn, axis=1)).flatten()) != 0:
            all_indices = np.arange(len(structure), dtype=int)
            constraint_xyz = np.argwhere(np.all(sel_dyn, axis=1)).flatten()
            not_constrained_xyz = np.setdiff1d(all_indices, constraint_xyz)
            # LAMMPS starts counting from 1
            constraint_xyz += 1
            ind_x = np.argwhere(sel_dyn[not_constrained_xyz, 0]).flatten()
            ind_y = np.argwhere(sel_dyn[not_constrained_xyz, 1]).flatten()
            ind_z = np.argwhere(sel_dyn[not_constrained_xyz, 2]).flatten()
            constraint_xy = not_constrained_xyz[np.intersect1d(ind_x, ind_y)] + 1
            constraint_yz = not_constrained_xyz[np.intersect1d(ind_y, ind_z)] + 1
            constraint_zx = not_constrained_xyz[np.intersect1d(ind_z, ind_x)] + 1
            constraint_x = (
                not_constrained_xyz[np.setdiff1d(np.setdiff1d(ind_x, ind_y), ind_z)] + 1
            )
            constraint_y = (
                not_constrained_xyz[np.setdiff1d(np.setdiff1d(ind_y, ind_z), ind_x)] + 1
            )
            constraint_z = (
                not_constrained_xyz[np.setdiff1d(np.setdiff1d(ind_z, ind_x), ind_y)] + 1
            )
            control_dict = {}
            if len(constraint_xyz) > 0:
                control_dict["group constraintxyz"] = "id " + " ".join(
                    [str(ind) for ind in constraint_xyz]
                )
                control_dict["fix constraintxyz"] = "constraintxyz setforce 0.0 0.0 0.0"
                if calc_md:
                    control_dict["velocity constraintxyz"] = "set 0.0 0.0 0.0"
            if len(constraint_xy) > 0:
                control_dict["group constraintxy"] = "id " + " ".join(
                    [str(ind) for ind in constraint_xy]
                )
                control_dict["fix constraintxy"] = "constraintxy setforce 0.0 0.0 NULL"
                if calc_md:
                    control_dict["velocity constraintxy"] = "set 0.0 0.0 NULL"
            if len(constraint_yz) > 0:
                control_dict["group constraintyz"] = "id " + " ".join(
                    [str(ind) for ind in constraint_yz]
                )
                control_dict["fix constraintyz"] = "constraintyz setforce NULL 0.0 0.0"
                if calc_md:
                    control_dict["velocity constraintyz"] = "set NULL 0.0 0.0"
            if len(constraint_zx) > 0:
                control_dict["group constraintxz"] = "id " + " ".join(
                    [str(ind) for ind in constraint_zx]
                )
                control_dict["fix constraintxz"] = "constraintxz setforce 0.0 NULL 0.0"
                if calc_md:
                    control_dict["velocity constraintxz"] = "set 0.0 NULL 0.0"
            if len(constraint_x) > 0:
                control_dict["group constraintx"] = "id " + " ".join(
                    [str(ind) for ind in constraint_x]
                )
                control_dict["fix constraintx"] = "constraintx setforce 0.0 NULL NULL"
                if calc_md:
                    control_dict["velocity constraintx"] = "set 0.0 NULL NULL"
            if len(constraint_y) > 0:
                control_dict["group constrainty"] = "id " + " ".join(
                    [str(ind) for ind in constraint_y]
                )
                control_dict["fix constrainty"] = "constrainty setforce NULL 0.0 NULL"
                if calc_md:
                    control_dict["velocity constrainty"] = "set NULL 0.0 NULL"
            if len(constraint_z) > 0:
                control_dict["group constraintz"] = "id " + " ".join(
                    [str(ind) for ind in constraint_z]
                )
                control_dict["fix constraintz"] = "constraintz setforce NULL NULL 0.0"
                if calc_md:
                    control_dict["velocity constraintz"] = "set NULL NULL 0.0"
    return control_dict



# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import contextlib
import os
from concurrent.futures import Future
from typing import Optional

from executorlib import BaseExecutor, SingleNodeExecutor

__author__ = "Sarath Menon, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 28, 2020"


def init_function():
    from lammps import lammps  # noqa: PLC0415
    from mpi4py import MPI  # noqa: PLC0415

    from pylammpsmpi.mpi.lmpmpi import select_cmd  # noqa: PLC0415

    class ParallelLammps:
        def __init__(self):
            self._job = None

        def start(self, argument_lst):
            args = []
            if len(argument_lst) > 0:
                args.extend(argument_lst)
            self._job = lammps(cmdargs=args)

        def shutdown(self):
            if self._job is not None:
                self._job.close()
                self._job = None

        def command(self, input_dict):
            output = select_cmd(input_dict["command"])(
                job=self._job, funct_args=input_dict["args"]
            )
            if MPI.COMM_WORLD.rank == 0:
                return output
            else:
                return True

    return {"lmp": ParallelLammps()}


def call_function_lmp(lmp, input_dict):
    return lmp.command(input_dict=input_dict)


def start_lmp(lmp, argument_lst):
    lmp.start(argument_lst=argument_lst)
    return True


def shutdown_lmp(lmp):
    lmp.shutdown()
    return True


class LammpsConcurrent:
    def __init__(
        self,
        cores: int = 8,
        oversubscribe: bool = False,
        working_directory: str = ".",
        hostname_localhost: Optional[bool] = None,
        cmdargs: list = None,
        executor: Optional[BaseExecutor] = None,
    ):
        """
        Initialize the LammpsConcurrent object.

        Parameters
        ----------
        cores : int, optional
            Number of cores to use for parallel execution (default is 8).
        oversubscribe : bool, optional
            Whether to oversubscribe the cores (default is False).
        working_directory : str, optional
            Working directory for Lammps execution (default is current directory).
        cmdargs : list, optional
            Additional command line arguments for Lammps (default is None).
        executor: Executor to use for parallel execution (default: None)

        Returns
        -------
        None
        """
        self.cores = cores
        self.working_directory = working_directory
        if executor is None:
            self._exe = SingleNodeExecutor(
                block_allocation=True,
                max_workers=1,
                init_function=init_function,
                hostname_localhost=hostname_localhost,
                openmpi_oversubscribe=oversubscribe,
                resource_dict={
                    "cores": self.cores,
                    "cwd": self.working_directory,
                },
            )
            self._external_executor = False
        else:
            self._exe = executor
            self._external_executor = True
        if cmdargs is None:
            cmdargs = []
        self._exe.submit(start_lmp, argument_lst=cmdargs).result()

    def _send_and_receive_dict(self, command, data=None):
        return self._exe.submit(
            call_function_lmp, input_dict={"command": command, "args": data}
        )

    @property
    def version(self) -> Future:
        """
        Get the version of Lammps.

        Returns
        -------
        version : Future
            A future object representing the version string of Lammps.
        """
        return self._send_and_receive_dict(command="get_version", data=[])

    def file(self, inputfile: str) -> Future:
        """
        Read script from an input file.

        Parameters
        ----------
        inputfile : str
            Name of the input file.

        Returns
        -------
        None
        """
        if not os.path.exists(inputfile):
            raise FileNotFoundError("Input file does not exist")
        return self._send_and_receive_dict(command="get_file", data=[inputfile])

    # TODO
    def extract_setting(self, *args):
        return self._send_and_receive_dict(command="extract_setting", data=list(args))

    def extract_global(self, name: str) -> Future:
        """
        Extract value of global simulation parameters.

        Parameters
        ----------
        name : str
            Name of the global simulation parameter.

        Returns
        -------
        value : Future
            A future object representing the value of the requested global parameter.
        """
        return self._send_and_receive_dict(command="extract_global", data=[name])

    def extract_box(self) -> Future:
        """
        Get the simulation box.

        Returns
        -------
        box : Future
            A future object representing the simulation box.
        """
        return self._send_and_receive_dict(command="extract_box", data=[])

    def extract_atom(self, name: str) -> Future:
        """
        Extract a property of the atoms.

        Parameters
        ----------
        name : str
            The property of the atom to be extracted.

        Returns
        -------
        val : Future
            A future object representing the extracted property of the atoms.
        """
        return self._send_and_receive_dict(command="extract_atom", data=[name])

    def extract_fix(self, *args) -> Future:
        """
        Extract a fix value.

        Parameters
        ----------
        args : tuple
            Variable number of arguments specifying the fix value to extract.

        Returns
        -------
        value : Future
            A future object representing the extracted fix value.
        """
        return self._send_and_receive_dict(command="extract_fix", data=list(args))

    def extract_variable(self, *args) -> Future:
        """
        Extract the value of a variable.

        Parameters
        ----------
        args : tuple
            Variable number of arguments specifying the variable to extract.

        Returns
        -------
        data : Future
            A future object representing the value of the variable.
        """
        return self._send_and_receive_dict(command="extract_variable", data=list(args))

    @property
    def natoms(self) -> Future:
        """
        Get the number of atoms.

        Returns
        -------
        natoms : Future
            A future object representing the number of atoms.
        """
        return self.get_natoms()

    def get_natoms(self) -> Future:
        """
        Get the number of atoms.

        Returns
        -------
        natoms : Future
            A future object representing the number of atoms.
        """
        return self._send_and_receive_dict(command="get_natoms", data=[])

    def set_variable(self, *args) -> Future:
        """
        Set the value of a string style variable.

        Parameters
        ----------
        args : tuple
            Variable number of arguments specifying the variable name and value.

        Returns
        -------
        flag : Future
            A future object representing the success flag (0 if successful, -1 otherwise).
        """
        return self._send_and_receive_dict(command="set_variable", data=list(args))

    def reset_box(self, *args) -> Future:
        """
        Reset the simulation box.

        Parameters
        ----------
        args : tuple
            Variable number of arguments specifying the box parameters.

        Returns
        -------
        None
        """
        return self._send_and_receive_dict(command="reset_box", data=list(args))

    def generate_atoms(
        self, ids=None, type=None, x=None, v=None, image=None, shrinkexceed=False
    ) -> Future:
        """
        Create atoms on all processors.

        Parameters
        ----------
        ids : list of ints, optional
            IDs of N atoms that need to be created.
        type : list of atom types, optional
            Type of N atoms.
        x : list of positions
            List of positions for N atoms.
        v : list of velocities, optional
            List of velocities for N atoms.
        image : list of ints, optional
            List of image flags for N atoms.
        shrinkexceed : bool, optional
            Whether to shrink the box if the atoms exceed the box boundaries.

        Returns
        -------
        None
        """
        return self.create_atoms(
            ids=ids, type=type, x=x, v=v, image=image, shrinkexceed=shrinkexceed
        )

    def create_atoms(
        self, n, id, type, x, v=None, image=None, shrinkexceed=False
    ) -> Future:
        """
        Create atoms on all processors.

        Parameters
        ----------
        n : int
            Number of atoms.
        id : list of ints, optional
            IDs of N atoms that need to be created.
        type : list of atom types, optional
            Type of N atoms.
        x : list of positions
            List of positions for N atoms.
        v : list of velocities, optional
            List of velocities for N atoms.
        image : list of ints, optional
            List of image flags for N atoms.
        shrinkexceed : bool, optional
            Whether to shrink the box if the atoms exceed the box boundaries.

        Returns
        -------
        None
        """
        if x is not None:
            funct_args = [n, id, type, x, v, image, shrinkexceed]
            return self._send_and_receive_dict(command="create_atoms", data=funct_args)
        else:
            raise TypeError("Value of x cannot be None")

    @property
    def has_exceptions(self) -> Future:
        """
        Return whether the Lammps shared library was compiled with C++ exceptions handling enabled.

        Returns
        -------
        has_exceptions : Future
            A future object representing whether the Lammps library has exceptions handling enabled.
        """
        return self._send_and_receive_dict(command="has_exceptions", data=[])

    @property
    def has_gzip_support(self) -> Future:
        """
        Return whether the Lammps shared library has gzip support.

        Returns
        -------
        has_gzip_support : Future
            A future object representing whether the Lammps library has gzip support.
        """
        return self._send_and_receive_dict(command="has_gzip_support", data=[])

    @property
    def has_png_support(self) -> Future:
        """
        Return whether the Lammps shared library has PNG support.

        Returns
        -------
        has_png_support : Future
            A future object representing whether the Lammps library has PNG support.
        """
        return self._send_and_receive_dict(command="has_png_support", data=[])

    @property
    def has_jpeg_support(self) -> Future:
        """
        Return whether the Lammps shared library has JPEG support.

        Returns
        -------
        has_jpeg_support : Future
            A future object representing whether the Lammps library has JPEG support.
        """
        return self._send_and_receive_dict(command="has_jpeg_support", data=[])

    @property
    def has_ffmpeg_support(self):
        return self._send_and_receive_dict(command="has_ffmpeg_support", data=[])

    @property
    def installed_packages(self):
        return self._send_and_receive_dict(command="get_installed_packages", data=[])

    def set_fix_external_callback(self, *args):
        """
        Follows the signature of LAMMPS's set_fix_external_callback(fix_id, callback, caller=None).
        The `caller` argument is a list of inputs to the callback function, including class instances and, importantly,
        the LammpsLibrary instance itself. Since LAMMPS's set_fix_external_callback allows passing to itself a
        self-reference, we want to keep this feature. Because the LAMMPS instance is unavailable at this stage, we
        insert a placeholder in its position, which will later be replaced with the actual LAMMPS instance.
        """

        def check_type(variable):
            if type(variable).__name__ in [
                "LammpsLibrary",
                "LammpsConcurrent",
                "LammpsBase",
            ]:
                return "pylammpsmpi.lammps.reference"
            else:
                return variable

        args = list(args)
        if len(args) == 3 and args[2]:
            if isinstance(args[2], list):
                args[2] = [check_type(variable=a) for a in args[2]]
            elif isinstance(args[2], dict):
                args[2] = {k: check_type(variable=v) for k, v in args[2].items()}
            elif isinstance(check_type(variable=args[2]), str):
                args[2] = "pylammpsmpi.lammps.reference"
            else:
                raise TypeError(f"Argument type {type(args[2])} not supported")
        return self._send_and_receive_dict(
            command="set_fix_external_callback", data=args
        )

    def get_neighlist(self, *args):
        """Returns an instance of :class:`NeighList` which wraps access to the neighbor list with the given index
        :param idx: index of neighbor list
        :type  idx: int
        :return: an instance of :class:`NeighList` wrapping access to neighbor list data
        :rtype:  NeighList
        """
        return self._send_and_receive_dict(command="get_neighlist", data=list(args))

    def find_pair_neighlist(self, *args):
        """Find neighbor list index of pair style neighbor list
        Try finding pair instance that matches style. If exact is set, the pair must
        match style exactly. If exact is 0, style must only be contained. If pair is
        of style pair/hybrid, style is instead matched the nsub-th hybrid sub-style.
        Once the pair instance has been identified, multiple neighbor list requests
        may be found. Every neighbor list is uniquely identified by its request
        index. Thus, providing this request index ensures that the correct neighbor
        list index is returned.
        :param style: name of pair style that should be searched for
        :type  style: string
        :param exact: controls whether style should match exactly or only must be contained in pair style name, defaults to True
        :type  exact: bool, optional
        :param nsub:  match nsub-th hybrid sub-style, defaults to 0
        :type  nsub:  int, optional
        :param request:   index of neighbor list request, in case there are more than one, defaults to 0
        :type  request:   int, optional
        :return: neighbor list index if found, otherwise -1
        :rtype:  int
        """
        return self._send_and_receive_dict(
            command="find_pair_neighlist", data=list(args)
        )

    def find_fix_neighlist(self, *args):
        """Find neighbor list index of fix neighbor list
        :param fixid: name of fix
        :type  fixid: string
        :param request:   index of neighbor list request, in case there are more than one, defaults to 0
        :type  request:   int, optional
        :return: neighbor list index if found, otherwise -1
        :rtype:  int
        """
        return self._send_and_receive_dict(
            command="find_fix_neighlist", data=list(args)
        )

    def find_compute_neighlist(self, *args):
        """Find neighbor list index of compute neighbor list
        :param computeid: name of compute
        :type  computeid: string
        :param request:   index of neighbor list request, in case there are more than one, defaults to 0
        :type  request:   int, optional
        :return: neighbor list index if found, otherwise -1
        :rtype:  int
        """
        return self._send_and_receive_dict(
            command="find_compute_neighlist", data=list(args)
        )

    def get_neighlist_size(self, *args):
        """Return the number of elements in neighbor list with the given index
        :param idx: neighbor list index
        :type  idx: int
        :return: number of elements in neighbor list with index idx
        :rtype:  int
        """
        return self._send_and_receive_dict(
            command="get_neighlist_size", data=list(args)
        )

    def get_neighlist_element_neighbors(self, *args):
        return self._send_and_receive_dict(
            command="get_neighlist_element_neighbors", data=list(args)
        )

    def command(self, cmd):
        """
        Send a command to the lammps object

        Parameters
        ----------
        cmd : string, list of strings
            command to be sent

        Returns
        -------
        None
        """
        if isinstance(cmd, list):
            return self._send_and_receive_dict(
                command="commands_string", data="\n".join(cmd)
            )
        elif len(cmd.split("\n")) > 1:
            return self._send_and_receive_dict(command="commands_string", data=cmd)
        else:
            return self._send_and_receive_dict(command="command", data=cmd)

    def gather_atoms(self, *args, concat=False, ids=None):
        """
        Gather atom properties

        Parameters
        ----------
        name : {'x', 'mass', 'id', 'type', 'mask', 'v', 'f',
                'molecule', 'q', 'mu', 'omega', 'angmom', 'torque', 'radius'}
            the property of atom to be extracted

        concat : bool, optional. Default False
            If True, gather information from all processors,
            but not sorted according to Atom ids

        ids : list, optional. Default None
            If a list of ids are provided, the required information
            for only those atoms are returned

        Returns
        -------
        val : array of length n_atoms sorted by atom ids
            If the requested name has multiple dimensions, output
            will be a multi-dimensional array.

        Notes
        -----
        This method gathers information from all processors.

        See Also
        --------
        extract_atoms
        """
        if concat:
            return self._send_and_receive_dict(
                command="gather_atoms_concat", data=list(args)
            )
        elif ids is not None:
            args = list(args)
            args.append(len(ids))
            args.append(ids)
            return self._send_and_receive_dict(command="gather_atoms_subset", data=args)
        else:
            return self._send_and_receive_dict(command="gather_atoms", data=list(args))

    def scatter_atoms(self, *args, ids=None):
        """
        Scatter atoms for the lammps library

        Args:
            *args:
        """
        if ids is not None:
            args = list(args)
            args.append(len(ids))
            args.append(ids)
            return self._send_and_receive_dict(
                command="scatter_atoms_subset", data=args
            )
        else:
            return self._send_and_receive_dict(command="scatter_atoms", data=list(args))

    def get_thermo(self, *args):
        """
        Return current value of thermo keyword

        Parameters
        ----------
        name : string
            name of the thermo keyword

        Returns
        -------
        val
            value of the thermo keyword

        """
        return self._send_and_receive_dict(command="get_thermo", data=list(args))

    # TODO
    def extract_compute(self, id, style, type, length=0, width=0):
        """
        Extract compute value from the lammps library

        Parameters
        ----------
        id : string
            id of the compute

        style: {0, 1}
            0 - global data
            1 - per atom data

        type: {0, 1, 2}
            0 - scalar
            1 - vector
            2 - array

        length: int, optional. Default 0
            if `style` is 0 and `type` is 1 or 2, then `length` is the length
            of vector.

        width: int, optional. Default 0
            if `type` is 2, then `width` is the number of elements in each
            element along length.

        Returns
        -------
        val
            data computed by the fix depending on the chosen inputs

        """
        args = [id, style, type, length, width]
        return self._send_and_receive_dict(command="extract_compute", data=args)

    def close(self):
        """
        Close the current lammps object

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        with contextlib.suppress(AttributeError):
            self._exe.submit(shutdown_lmp).result()
        if not self._external_executor:
            self._exe.shutdown(wait=True)
        self._exe = None

    # TODO
    def __del__(self):
        if self._exe is not None:
            self.close()



# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from concurrent.futures import Future
from typing import Any, Union

from pylammpsmpi.wrapper.concurrent import LammpsConcurrent

__author__ = "Sarath Menon, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Feb 28, 2020"


def get_result(future: Future, cores: int) -> Any:
    if cores > 1:
        return future.result()[0]
    else:
        return future.result()


class LammpsBase(LammpsConcurrent):
    @property
    def version(self) -> str:
        """
        Get the version of lammps

        Returns:
        version: str
            version string of lammps
        """
        return get_result(future=super().version, cores=self.cores)

    def file(self, inputfile: str) -> None:
        """
        Read script from an input file

        Parameters:
        inputfile: str
            name of inputfile

        Returns:
        None
        """
        _ = super().file(inputfile=inputfile).result()

    # TODO
    def extract_setting(self, *args) -> Union[int, float, str]:
        """
        Extract a setting value

        Parameters:
        *args: tuple
            arguments to specify the setting to extract

        Returns:
        value: int, float, or str
            extracted setting value
        """
        return get_result(future=super().extract_setting(*args), cores=self.cores)

    def extract_global(self, name: str) -> Union[int, float, str]:
        """
        Extract value of global simulation parameters

        Parameters:
        name : str
            name of the global parameter to extract

        Returns:
        value: int, float, or str
            extracted value of the global parameter
        """
        return get_result(future=super().extract_global(name=name), cores=self.cores)

    def extract_box(self) -> list[Union[float, list[float], list[int]]]:
        """
        Get the simulation box

        Returns:
        box : list
            list containing the simulation box information
            [boxlo, boxhi, xy, yz, xz, periodicity, box_change]
            where boxlo and boxhi are lower and upper bounds of the box in three dimensions,
            xy, yz, xz are the box tilts, periodicity is an array which shows if
            the box is periodic in three dimensions, and box_change is a list of booleans
            indicating if the box dimensions have changed
        """
        return get_result(future=super().extract_box(), cores=self.cores)

    def extract_atom(self, name: str) -> Union[list[int], list[float]]:
        """
        Extract a property of the atoms

        Parameters:
        name : str
            the property of atom to be extracted

        Returns:
        val : list of int or float
            If the requested name has multiple dimensions, output
            will be a multi-dimensional list.
        """
        return get_result(future=super().extract_atom(name=name), cores=self.cores)

    def extract_fix(self, *args) -> Union[int, float, list[Union[int, float]]]:
        """
        Extract a fix value

        Parameters:
        *args: tuple
            arguments to specify the fix value to extract

        Returns:
        value: int, float, or list of int or float
            extracted fix value corresponding to the requested dimensions
        """
        return get_result(future=super().extract_fix(*args), cores=self.cores)

    def extract_variable(self, *args) -> Union[int, float, list[Union[int, float]]]:
        """
        Extract the value of a variable

        Parameters:
        *args: tuple
            arguments to specify the variable to extract

        Returns:
        data: int, float, or list of int or float
            value of the variable depending on the requested dimension
        """
        return get_result(future=super().extract_variable(*args), cores=self.cores)

    @property
    def natoms(self) -> int:
        """
        Get the number of atoms

        Returns:
        natoms : int
            number of atoms
        """
        return self.get_natoms()

    def get_natoms(self) -> int:
        """
        Get the number of atoms

        Returns:
        natoms : int
            number of atoms
        """
        return get_result(future=super().get_natoms(), cores=self.cores)

    def set_variable(self, *args) -> int:
        """
        Set the value of a string style variable

        Parameters:
        *args: tuple
            arguments to specify the variable and its value

        Returns:
        flag : int
            0 if successful, -1 otherwise
        """
        return get_result(future=super().set_variable(*args), cores=self.cores)

    def reset_box(self, *args) -> None:
        """
        Reset the simulation box

        Parameters:
        *args: tuple
            arguments to specify the new box dimensions

        Returns:
        None
        """
        _ = super().reset_box(*args).result()

    def generate_atoms(
        self,
        ids: list[int] = None,
        type: list[int] = None,
        x: list[float] = None,
        v: list[float] = None,
        image: list[int] = None,
        shrinkexceed: bool = False,
    ) -> None:
        """
        Create atoms on all procs

        Parameters:
        ids : list of ints, optional
            ids of N atoms that need to be created
            if not specified, ids from 1 to N are assigned

        type : list of atom types, optional
            type of N atoms, if not specified, all atoms are assigned as type 1

        x: list of positions
            list of the type `[posx, posy, posz]` for N atoms

        v: list of velocities
            list of the type `[vx, vy, vz]` for N atoms

        image: list of ints, optional
            if not specified a list of 0s will be used.

        shrinkexceed: bool, optional
            default False

        Returns:
        None
        """
        _ = (
            super()
            .generate_atoms(
                ids=ids, type=type, x=x, v=v, image=image, shrinkexceed=shrinkexceed
            )
            .result()
        )

    def create_atoms(
        self,
        n: int,
        id: list[int],
        type: list[int],
        x: list[float],
        v: list[float] = None,
        image: list[int] = None,
        shrinkexceed: bool = False,
    ) -> None:
        """
        Create atoms on all procs

        Parameters:
        n : int
            number of atoms

        id : list of ints, optional
            ids of N atoms that need to be created
            if not specified, ids from 1 to N are assigned

        type : list of atom types, optional
            type of N atoms, if not specified, all atoms are assigned as type 1

        x: list of positions
            list of the type `[posx, posy, posz]` for N atoms

        v: list of velocities
            list of the type `[vx, vy, vz]` for N atoms

        image: list of ints, optional
            if not specified a list of 0s will be used.

        shrinkexceed: bool, optional
            default False

        Returns:
        None
        """
        _ = (
            super()
            .create_atoms(
                n=n, id=id, type=type, x=x, v=v, image=image, shrinkexceed=shrinkexceed
            )
            .result()
        )

    @property
    def has_exceptions(self) -> bool:
        """Return whether the LAMMPS shared library was compiled with C++ exceptions handling enabled"""
        return get_result(future=super().has_exceptions, cores=self.cores)

    @property
    def has_gzip_support(self) -> bool:
        return get_result(future=super().has_gzip_support, cores=self.cores)

    @property
    def has_png_support(self) -> bool:
        return get_result(future=super().has_png_support, cores=self.cores)

    @property
    def has_jpeg_support(self) -> bool:
        return get_result(future=super().has_jpeg_support, cores=self.cores)

    @property
    def has_ffmpeg_support(self) -> bool:
        return get_result(future=super().has_ffmpeg_support, cores=self.cores)

    @property
    def installed_packages(self) -> list[str]:
        return get_result(future=super().installed_packages, cores=self.cores)

    def set_fix_external_callback(self, *args) -> None:
        _ = super().set_fix_external_callback(*args).result()

    def get_neighlist(self, *args):
        """Returns an instance of :class:`NeighList` which wraps access to the neighbor list with the given index
        :param idx: index of neighbor list
        :type  idx: int
        :return: an instance of :class:`NeighList` wrapping access to neighbor list data
        :rtype:  NeighList
        """
        return get_result(future=super().get_neighlist(*args), cores=self.cores)

    def find_pair_neighlist(self, *args) -> int:
        """Find neighbor list index of pair style neighbor list
        Try finding pair instance that matches style. If exact is set, the pair must
        match style exactly. If exact is 0, style must only be contained. If pair is
        of style pair/hybrid, style is instead matched the nsub-th hybrid sub-style.
        Once the pair instance has been identified, multiple neighbor list requests
        may be found. Every neighbor list is uniquely identified by its request
        index. Thus, providing this request index ensures that the correct neighbor
        list index is returned.
        :param style: name of pair style that should be searched for
        :type  style: string
        :param exact: controls whether style should match exactly or only must be contained in pair style name, defaults to True
        :type  exact: bool, optional
        :param nsub:  match nsub-th hybrid sub-style, defaults to 0
        :type  nsub:  int, optional
        :param request:   index of neighbor list request, in case there are more than one, defaults to 0
        :type  request:   int, optional
        :return: neighbor list index if found, otherwise -1
        :rtype:  int
        """
        return get_result(future=super().find_pair_neighlist(*args), cores=self.cores)

    def find_fix_neighlist(self, *args):
        """Find neighbor list index of fix neighbor list
        :param fixid: name of fix
        :type  fixid: string
        :param request:   index of neighbor list request, in case there are more than one, defaults to 0
        :type  request:   int, optional
        :return: neighbor list index if found, otherwise -1
        :rtype:  int
        """
        return get_result(future=super().find_fix_neighlist(*args), cores=self.cores)

    def find_compute_neighlist(self, *args):
        """Find neighbor list index of compute neighbor list
        :param computeid: name of compute
        :type  computeid: string
        :param request:   index of neighbor list request, in case there are more than one, defaults to 0
        :type  request:   int, optional
        :return: neighbor list index if found, otherwise -1
        :rtype:  int
        """
        return get_result(
            future=super().find_compute_neighlist(*args), cores=self.cores
        )

    def get_neighlist_size(self, *args):
        """Return the number of elements in neighbor list with the given index
        :param idx: neighbor list index
        :type  idx: int
        :return: number of elements in neighbor list with index idx
        :rtype:  int
        """
        return get_result(future=super().get_neighlist_size(*args), cores=self.cores)

    def get_neighlist_element_neighbors(self, *args):
        return get_result(
            future=super().get_neighlist_element_neighbors(*args), cores=self.cores
        )

    def command(self, cmd):
        """
        Send a command to the lammps object

        Parameters
        ----------
        cmd : string, list of strings
            command to be sent

        Returns
        -------
        None
        """
        _ = super().command(cmd=cmd).result()

    def gather_atoms(self, *args, concat=False, ids=None):
        """
        Gather atom properties

        Parameters
        ----------
        name : {'x', 'mass', 'id', 'type', 'mask', 'v', 'f',
                'molecule', 'q', 'mu', 'omega', 'angmom', 'torque', 'radius'}
            the property of atom to be extracted

        concat : bool, optional. Default False
            If True, gather information from all processors,
            but not sorted according to Atom ids

        ids : list, optional. Default None
            If a list of ids are provided, the required information
            for only those atoms are returned

        Returns
        -------
        val : array of length n_atoms sorted by atom ids
            If the requested name has multiple dimensions, output
            will be a multi-dimensional array.

        Notes
        -----
        This method gathers information from all processors.

        See Also
        --------
        extract_atoms
        """
        return get_result(
            future=super().gather_atoms(*args, concat=concat, ids=ids), cores=self.cores
        )

    def scatter_atoms(self, *args, ids=None):
        """
        Scatter atoms for the lammps library

        Args:
            *args:
        """
        _ = super().scatter_atoms(*args, ids=ids).result()

    def get_thermo(self, *args):
        """
        Return current value of thermo keyword

        Parameters
        ----------
        name : string
            name of the thermo keyword

        Returns
        -------
        val
            value of the thermo keyword

        """
        return get_result(future=super().get_thermo(*args), cores=self.cores)

    # TODO
    def extract_compute(self, id, style, type, length=0, width=0):
        """
        Extract compute value from the lammps library

        Parameters
        ----------
        id : string
            id of the compute

        style: {0, 1}
            0 - global data
            1 - per atom data

        type: {0, 1, 2}
            0 - scalar
            1 - vector
            2 - array

        length: int, optional. Default 0
            if `style` is 0 and `type` is 1 or 2, then `length` is the length
            of vector.

        width: int, optional. Default 0
            if `type` is 2, then `width` is the number of elements in each
            element along length.

        Returns
        -------
        val
            data computed by the fix depending on the chosen inputs

        """
        return get_result(
            future=super().extract_compute(
                id=id, style=style, type=type, length=length, width=width
            ),
            cores=self.cores,
        )



import pylammpsmpi._version
from pylammpsmpi.wrapper.base import LammpsBase
from pylammpsmpi.wrapper.concurrent import LammpsConcurrent, init_function
from pylammpsmpi.wrapper.extended import LammpsLibrary

__all__ = ["LammpsLibrary", "LammpsConcurrent", "LammpsBase", "init_function"]
__version__ = pylammpsmpi._version.__version__


try:
    from pylammpsmpi.wrapper.ase import LammpsASELibrary

    __all__ += ["LammpsASELibrary"]
except ImportError:
    pass
